package hr.fer.phd.radovan;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.TimeZone;

import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.Point;
import org.opencv.core.RotatedRect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgcodecs.Imgcodecs;
import org.opencv.imgproc.Imgproc;
import org.opencv.imgproc.Moments;

/**
 * Generates data for irradiance estimation curve based on sun and cloud
 * tracking data. Wide-angle ground-based camera GoPro Fusion was used to obtain
 * images of the sky.
 * 
 * @author Aleksander Radovan
 *
 */
public class IrradianceEstimationCurveGenerator {

	// data structures for saving contours and centroid points of clouds and the sun
	private static List<List<Point>> cloudCentroids = new ArrayList<>();
	private static List<List<MatOfPoint>> cloudContours = new ArrayList<>();
	private static List<Point> sunCentroids = new ArrayList<>();

	/**
	 * Starts the algorithm for generating the data for irradiance estimation curve.
	 * 
	 * @param args - command line arguments (not used)
	 * @throws FileNotFoundException thrown if sky images not found
	 */
	public static void main(String[] args) throws FileNotFoundException {

		// fetching the start time for measuring time duration for the algorithm
		long start = System.currentTimeMillis();

		// loading OpenCV DLL files for Windows operating system,
		// VM parameter needs to be set:
		// (-Djava.library.path=C:\DEV\LIBs\OpenCV-4.3.0\opencv\build\java\x64)
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

		// data structure for saving coverage records
		List<List<CoverageRecord>> data = new ArrayList<>();

		// loop that processes all ordinal numbers of folders with ordinal numbers sky
		// images
		// created by the GoPro Fusion wide-angle ground-based camera
		for (int i = 205; i <= 205; i++) {

			// creating original image path to get information about the original time
			// of taking the image
			String originalmagePath = "D:\\Fusion\\15-07-2020\\" + i + "GFRNT\\";

			// creating the image path of undirstorted image
			String imagePath = "D:\\Fusion\\15-07-2020\\" + i + "GFRNT\\undistorted";

			// fetching all the files from the folder of undistorted images
			File[] listOfFiles = new File(imagePath).listFiles();

			// declaring the variable that will contain the last knows sun centroid point
			// in case of cloud covering the sun
			Point lastKnownSunCentroid = null;

			// image ordinal number
			long imageNumber = 0;

			// loop that fetches all the image files inside of the current folder
			for (File imageFile : listOfFiles) {

				// if the file system item is not an image (in case of nested folder),
				// skip it
				if (imageFile.isFile() == false) {
					continue;
				}

				// fetch the image number from the image name created by the camera (for
				// example, "GF018382.JPG"),
				// the number between the third (index 2) and the ninth (index 8) is being
				// extracted and parsed
				Integer imageOrdinalNumber = Integer.parseInt(imageFile.getName().substring(2, 8));

				// original file that represents the sky image with the full path
				File originalImageFile = new File(originalmagePath + "GF0" + imageOrdinalNumber + ".JPG");

				// printing out the current image name being processed
				System.out.println("Processing " + imagePath + "\\" + imageFile.getName());

				// getting the time when the original image was taken from the camera
				LocalDateTime triggerTime = LocalDateTime.ofInstant(
						Instant.ofEpochMilli(originalImageFile.lastModified()), TimeZone.getDefault().toZoneId());

				// printing out the current time
				System.out.println(triggerTime);

				// creating the object for the centroid of the sun
				Point sunCentroid = new Point();

				// creating the OpenCV image container for detecting the sun on the sky
				Mat originalSkyImageMat = Imgcodecs.imread(imageFile.getAbsolutePath());

				// creating the OpenCV image container for detecting clouds on the sky
				Mat originalCloudsImageMat = Imgcodecs.imread(imageFile.getAbsolutePath());

				// creating the OpenCV image container for HSV color model of the image with the
				// sun on the sky
				Mat hsvSkyImage = new Mat();

				// creating the OpenCV image container for HSV color model of the image with
				// clouds on the sky
				Mat hsvCloudsImage = new Mat();

				// creating the OpenCV image container for creating the mask for the shape of
				// the sun
				Mat maskForSun = new Mat();

				// creating the OpenCV image container for creating the mask for the shape of
				// clouds
				Mat maskForClouds = new Mat();

				// converting the image with the sky from RGB to HSV color model
				Imgproc.cvtColor(originalSkyImageMat, hsvSkyImage, Imgproc.COLOR_BGR2HSV);

				// converting the image with the sky from RGB to HSV color model
				Imgproc.cvtColor(originalCloudsImageMat, hsvCloudsImage, Imgproc.COLOR_BGR2HSV);

				// creating the OpenCV image container for the results of mophological operation
				// erosion on the image with the sky
				Mat morphOutputForSky = new Mat();

				// creating the OpenCV image container for the results of mophological operation
				// erosion and dilation on the image with clouds
				Mat morphOutputForClouds = new Mat();

				// setting the minimum and maximum values for conversion applying threshold
				// filters on images with the sun and the clouds
				Scalar minValuesForSun = new Scalar(0, 0, 240);
				Scalar maxValuesForSun = new Scalar(0, 0, 255);
				Scalar minValuesForClouds = new Scalar(0, 0, 0);
				Scalar maxValuesForClouds = new Scalar(70, 70, 255);

				// filtering the HSV images with defined scalar thresholds
				Core.inRange(hsvSkyImage, minValuesForSun, maxValuesForSun, maskForSun);
				Core.inRange(hsvCloudsImage, minValuesForClouds, maxValuesForClouds, maskForClouds);

				// applying morphological operations of erosion and dilation to images with the
				// sun and clouds
				// structural element size for eroding the image of the sky is 30 x 30
				// structural element size for eroding the image with clouds is 20 x 20
				// structural element size for dilating the image with clouds is 30 x 30
				Mat dilateElementForClouds = Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(30, 30));
				Mat erodeElementForSky = Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(50, 50));
				Mat erodeElementForClouds = Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(20, 20));

				// applying the erosion operation to the image with the sun and the clouds
				Imgproc.erode(maskForSun, morphOutputForSky, erodeElementForSky);
				Imgproc.erode(maskForClouds, morphOutputForClouds, erodeElementForClouds);

				// applying the dilation operation to the image with the clouds
				Imgproc.dilate(maskForClouds, morphOutputForClouds, dilateElementForClouds);

				// creating a data structure for saving contours of the sun
				List<MatOfPoint> contoursForSun = new ArrayList<>();

				// creating the OpenCV image container for hierarchy levels of contours of the
				// sun
				Mat hierarchyForSky = new Mat();

				// finding countours of the sun
				Imgproc.findContours(morphOutputForSky, contoursForSun, hierarchyForSky, Imgproc.RETR_TREE,
						Imgproc.CHAIN_APPROX_SIMPLE);

				// creating a data structure for saving contours of clouds
				List<MatOfPoint> contoursForClouds = new ArrayList<>();

				// creating the OpenCV image container for hierarchy levels of contours of
				// clouds
				Mat hierarchyForClouds = new Mat();

				// finding countours of clouds
				Imgproc.findContours(morphOutputForClouds, contoursForClouds, hierarchyForClouds, Imgproc.RETR_TREE,
						Imgproc.CHAIN_APPROX_SIMPLE);

				// declaring the variables for sun contour area, sun rectangle area, sun contour
				// index in the list
				// of detected contours (if more shiny contours on the sky detected), detected
				// rectangle around
				// the sun and the flag if the sun has been detected or if it's covered with
				// clouds
				double sunArea = 0;
				double sunRectangleArea = 0;
				int sunContourIndex = 0;
				RotatedRect rotatedRectAroundSun = null;
				boolean sunDetected = false;

				// declaring four points for rectangle around the sun contour
				Point[] sunPoints = new Point[4];

				// check if the sun has been detected on the sky image
				if (contoursForSun.size() > 0) {

					// reseting the sun area to zero
					sunArea = 0;

					// loop that fetches all candidates for the sun and saves the
					// area and the contour index with the largest area (if more than
					// one shiny contours found on the image)
					for (int k = 0; k < contoursForSun.size(); k++) {
						MatOfPoint sunContour = contoursForSun.get(k);
						double newArea = Imgproc.contourArea(sunContour);
						if (newArea > sunArea) {
							sunArea = newArea;
							sunContourIndex = k;
						}
					}

					// drawing the contours of the sun on the image
					Imgproc.drawContours(originalSkyImageMat, contoursForSun, sunContourIndex, new Scalar(0, 0, 255),
							10, 2, hierarchyForSky);

					// calculating moments in order to determine the sun contour
					Moments moments = Imgproc.moments(contoursForSun.get(sunContourIndex));
					sunCentroid.x = moments.get_m10() / moments.get_m00();
					sunCentroid.y = moments.get_m01() / moments.get_m00();

					// saving the centroid point of the detected sun
					sunCentroids.add(sunCentroid);

					// if sun area has the minimum area of 5000 pixels, it is considered
					// as successfuly detected sun and the last known sun centroid is saved
					if (sunArea > 5000) {
						lastKnownSunCentroid = sunCentroid;
					}

					// setting the flag for successfuly detected sun if any of the contours are
					// found
					if (sunArea > 0) {
						sunDetected = true;
					}

					// drawing a circle that represents the centroid of the sun
					Imgproc.circle(originalSkyImageMat, sunCentroid, 10, new Scalar(0, 0, 255), 10);

					// determining the rotated rectangle shape around the contour of the sun
					rotatedRectAroundSun = Imgproc
							.minAreaRect(new MatOfPoint2f(contoursForSun.get(sunContourIndex).toArray()));

					// calculating the area of the rotated rectangle around the sun
					sunRectangleArea = rotatedRectAroundSun.boundingRect().area();

					// initialing the array of points for rectangle shape
					sunPoints = new Point[4];

					// setting the rectangle points around the sun contour
					rotatedRectAroundSun.points(sunPoints);

					// loop that draws the rectangle with four lines that connect four points
					// detected in previous steps
					for (int p = 0; p < 4; p++) {
						Imgproc.line(originalSkyImageMat, sunPoints[p], sunPoints[(p + 1) % 4], new Scalar(0, 0, 255),
								5);
					}

				} else {
					// if contours of the sun are not detected, the flag is set to false
					sunDetected = false;
				}

				// creating data structures for saving clouds centroids and contours
				List<Point> listOfCentroidsForImage = new ArrayList<>();
				List<MatOfPoint> listOfContours = new ArrayList<>();

				// checking if any clouds are detected
				if (contoursForClouds.size() > 0) {

					// loop for iterating through the cloud contours
					for (int k = 0; k < contoursForClouds.size(); k++) {

						// calculating area of the cloud contour
						double cloudArea = Imgproc.contourArea(contoursForClouds.get(k));

						// if the area has a significant value (at least 60% of the sun area)
						if (cloudArea > 3000) {

							// calculate the moments to determine centroid points of the cloud
							Moments cloudMoments = Imgproc.moments(contoursForClouds.get(k));
							Point cloudCentroid = new Point();
							cloudCentroid.x = cloudMoments.get_m10() / cloudMoments.get_m00();
							cloudCentroid.y = cloudMoments.get_m01() / cloudMoments.get_m00();

							// if the cloud position is outside the range of covering the sun - ignore it
							// (centroid of the cloud is on the possible path of the sun)
							if (cloudCentroid.y > 600 && cloudCentroid.y < 2500) {

								// if the cloud centroid has "passed" the sun position - ignore it
								if ((sunCentroid.x + 300) > cloudCentroid.x
										&& (sunCentroid.y + 300) > cloudCentroid.y) {

									// if the cloud centroid is too near the sun centroid - it's the sun's contour,
									// so skip it
									if (calculateDistance(sunCentroid, cloudCentroid) > 100) {
										// draw the contour of the cloud
										Imgproc.drawContours(originalCloudsImageMat, contoursForClouds, k,
												new Scalar(255, 0, 0), 10, 2, hierarchyForClouds);

										// calculate the bounding rectangle around the cloud
										RotatedRect rotatedRectAroundCloud = Imgproc
												.minAreaRect(new MatOfPoint2f(contoursForClouds.get(k).toArray()));

										// initialize the points that define the bouding rectangle around the cloud
										Point[] points = new Point[4];

										// set the rectangle points
										rotatedRectAroundCloud.points(points);

										// draw the rectangle with four lines
										for (int p = 0; p < 4; p++) {
											Imgproc.line(originalCloudsImageMat, points[p], points[(p + 1) % 4],
													new Scalar(0, 255, 0), 5);
										}

										listOfCentroidsForImage.add(cloudCentroid);
										listOfContours.add(contoursForClouds.get(k));
									}
								}
							}
						}
					}
				}

				// saving cloud contours and centroid points
				cloudCentroids.add(listOfCentroidsForImage);
				cloudContours.add(listOfContours);

				// if twelve images are processed, the simulation of the next period can start
				if (cloudContours.size() > 12) {
					
					// fetching the contours that were detected on the last image and 
					// a minute ago (12 x 5 seconds frequency of taking image = 60 seconds = 1 minute)
					List<MatOfPoint> firstPhaseContours = cloudContours.get(cloudContours.size() - 13);
					List<MatOfPoint> secondPhaseContours = cloudContours.get(cloudContours.size() - 1);

					// fetching the centroids that were detected on the last image and 
					// a minute ago
					Point firstSunCentroidPoint = sunCentroids.get(sunCentroids.size() - 13);
					Point secondSunCentroidPoint = sunCentroids.get(sunCentroids.size() - 1);

					// calculating the sun shift within one period of 5 seconds
					double sunShiftX = (secondSunCentroidPoint.x - firstSunCentroidPoint.x) / 12;
					double sunShiftY = (secondSunCentroidPoint.y - firstSunCentroidPoint.y) / 12;

					// drawing the vector of sun's movement
					Imgproc.circle(originalSkyImageMat, secondSunCentroidPoint, 1, new Scalar(0, 0, 255), 10);
					Imgproc.arrowedLine(originalSkyImageMat, firstSunCentroidPoint, secondSunCentroidPoint, new Scalar(0, 0, 255), 10, 3, 0);

					// initializing the second centroid index and point
					int secondContourIndex = 0;
					Point secondCentroid = new Point();

					// loops that finds the nearest centroid point of the cloud to determine the speed vector
					for (MatOfPoint secondContour : secondPhaseContours) {
						
						// calculating the centroid of the cloud
						Moments secondCentroidMoments = Imgproc.moments(secondContour);
						secondCentroid.x = secondCentroidMoments.get_m10() / secondCentroidMoments.get_m00();
						secondCentroid.y = secondCentroidMoments.get_m01() / secondCentroidMoments.get_m00();
						
						// initialization of the nearest centroid point
						Point nearestCentroid = new Point();
						
						// initializing the distance to the nearest contour variable
						double distanceToNearestContour = -1;
						
						// loop that finds the nearest contour
						for (MatOfPoint firstContour : firstPhaseContours) {
							
							// initializing and determining the centroid point of the contour
							Point firstCentroid = new Point();
							Moments firstCentroidMoments = Imgproc.moments(firstContour);
							firstCentroid.x = firstCentroidMoments.get_m10() / firstCentroidMoments.get_m00();
							firstCentroid.y = firstCentroidMoments.get_m01() / firstCentroidMoments.get_m00();
							
							// calculating new distance between the the first and the second cloud centroid
							double newDistance = calculateDistance(secondCentroid, firstCentroid);
							
							// if distance is positive, check if it's the new nearest distance and save it
							if (newDistance > 0) {
								if (distanceToNearestContour == -1) {
									distanceToNearestContour = newDistance;
									nearestCentroid = firstCentroid;
								} else if (distanceToNearestContour > newDistance) {
									distanceToNearestContour = newDistance;
									nearestCentroid = firstCentroid;
								}
							}
						}

						// calculate the cloud shift within five seconds
						double shiftX = (secondCentroid.x - nearestCentroid.x) / 12;
						double shiftY = (secondCentroid.y - nearestCentroid.y) / 12;

						// initialize the contour of the cloud
						MatOfPoint cloudContour = secondPhaseContours.get(secondContourIndex);

						// if the contour is not the sun's contour (the centroid needs to be at least 100 pixels away)
						// and it's still in the area of sun's path
						if (calculateDistance(sunCentroid,
								secondCentroid) > 100 && (secondCentroid.y > 300) && (secondCentroid.y < 2600)) {

							// creating the OpenCV image container for drawing the cloud path
							Mat cloudPathImage = Imgcodecs.imread(imageFile.getAbsolutePath());

							// determining rotated rectangle around the cloud
							RotatedRect rotatedRectAroundCloud = Imgproc
									.minAreaRect(new MatOfPoint2f(cloudContour.toArray()));

							// initializing and drawing the rectangle points of the cloud
							Point[] points = new Point[4];
							rotatedRectAroundCloud.points(points);
							for (int p = 0; p < 4; p++) {
								Imgproc.line(cloudPathImage, points[p], points[(p + 1) % 4], new Scalar(0, 255, 0), 5);
							}

							// drawing the rectangle points of the sun
							for (int p = 0; p < 4; p++) {
								Imgproc.line(cloudPathImage, sunPoints[p], sunPoints[(p + 1) % 4],
										new Scalar(0, 0, 255), 5);
							}

							// initialization of variables for tracking the upper left corner of the
							// bounding rectangle
							double min_x = 0;
							double min_y = 0;

							// fetching the original time of taking the sky image
							LocalTime cloudTime = LocalTime.ofInstant(
									Instant.ofEpochMilli(originalImageFile.lastModified()),
									TimeZone.getDefault().toZoneId());

							// initializing the data structure for sun coverage records
							List<CoverageRecord> coverageRecordList = new ArrayList<>();

							// declaring a variable that counts covered periods
							int periods = 0;

							// loop that simulates the movement of clouds and the sun in the future
							// checking the upper left corner of the bounding rectangle
							// if it's still not fully visible on the sky image, the coordinates are outside of the image (x >= -200, y >= -200)
							// if it's left the sky image area, x > 3104, y > 3000 (image resolution)
							while ((min_x >= -200 && min_x <= 3104) && (min_y >= -200 && min_y <= 3000)) {

								// creating the OpenCV image container for intersection region of the cloud and the sun
								Mat intersectingRegion = new Mat();

								// fetching the status of the intersection (can be "INTERSECT_FULL", "INTERSECT_PARTIAL" or "INTERSECT_NONE")
								int status = Imgproc.rotatedRectangleIntersection(rotatedRectAroundSun,
										rotatedRectAroundCloud, intersectingRegion);

								// declaring the variable that holds intersectionArea
								double intersectingArea = 0;

								// if the clouds rectangle intersects the sun's rectangle
								if (status == Imgproc.INTERSECT_FULL || status == Imgproc.INTERSECT_PARTIAL) {
									
									// calculate the intersection area
									intersectingArea = Imgproc.contourArea(intersectingRegion);
								}

								// add the coverage record to the list
								coverageRecordList.add(new CoverageRecord(secondContourIndex, cloudTime,
										intersectingArea / sunRectangleArea));

								// increase the cloud time that is a foundation for prediction
								// increase current cloud time by five seconds and increase
								// corresponding minutes and hours if needed
								if (cloudTime.getMinute() + 1 > 59) {
									if (cloudTime.getHour() + 1 < 24) {
										cloudTime = LocalTime.of(cloudTime.getHour() + 1, 0, cloudTime.getSecond());
									} else {
										cloudTime = LocalTime.of(0, 0, cloudTime.getSecond());
									}
								} else if (cloudTime.getSecond() + 5 > 59) {
									cloudTime = LocalTime.of(cloudTime.getHour(), cloudTime.getMinute() + 1,
											cloudTime.getSecond() - 55);
								} else {
									cloudTime = LocalTime.of(cloudTime.getHour(), cloudTime.getMinute(),
											cloudTime.getSecond() + 5);
								}

								// shifting the cloud's bounding rectangle by shiftX and shiftY amounts (in pixels)
								for (int p = 0; p < 4; p++) {
									points[p].x = points[p].x + shiftX;
									points[p].y = points[p].y + shiftY;
								}

								// shifting the sun's bounding rectangle by shiftX and shiftY amounts (in pixels)
								for (int p = 0; p < 4; p++) {
									sunPoints[p].x = sunPoints[p].x + sunShiftX;
									sunPoints[p].y = sunPoints[p].y + sunShiftY;
								}

								// determining the shifted rectangle position of the cloud
								MatOfPoint2f cloudRectanglePoints = new MatOfPoint2f(points[0], points[1], points[2],
										points[3]);
								
								// determining the area of the cloud's bounded rectangle
								rotatedRectAroundCloud = Imgproc.minAreaRect(cloudRectanglePoints);

								// determining the shifted rectangle position of the sun
								MatOfPoint2f sunRectanglePoints = new MatOfPoint2f(sunPoints[0], sunPoints[1],
										sunPoints[2], sunPoints[3]);
								
								// determining the area of the sun's bounded rectangle
								rotatedRectAroundSun = Imgproc.minAreaRect(sunRectanglePoints);

								// if the period variable hits a five minutes threshold (60 x 5 seconds = 300 seconds =  5 minutes)
								if (periods % 60 == 0) {
									
									// draw a rectangle of simulated position of the cloud in the future
									for (int p = 0; p < 4; p++) {
										Imgproc.line(cloudPathImage, points[p], points[(p + 1) % 4],
												new Scalar(0, 0, 0), 5);
									}
									
									// put time text to the drawn rectangle on the sky image
									Imgproc.putText(cloudPathImage,
											cloudTime.format(DateTimeFormatter.ofPattern("HH:mm:ss")),
											new Point(points[2].x, points[2].y - 50), 2, 2.0, new Scalar(0, 0, 0), 2);
								}

								// fetch the upper left corner of the bounding rectangle to determine if the cloud
								// needs to be considered as candidate for covering the sun
								min_x = Arrays.stream(points).mapToDouble(p -> p.x).sorted().min().getAsDouble();
								min_y = Arrays.stream(points).mapToDouble(p -> p.y).sorted().min().getAsDouble();

								// increase the number of periods
								periods++;
							}

							// add coverage record list of a cloud to a global data structure
							data.add(coverageRecordList);
						}

						// increasing the contour index to analyze the next cloud on the sky image
						secondContourIndex++;
					}
				}

				// releasing the memory reserved by image containers
				hierarchyForSky.release();
				hierarchyForClouds.release();
				erodeElementForSky.release();
				erodeElementForClouds.release();
				dilateElementForClouds.release();
				morphOutputForSky.release();
				morphOutputForClouds.release();
				maskForSun.release();
				maskForClouds.release();
				hsvSkyImage.release();
				hsvCloudsImage.release();
				originalSkyImageMat.release();
				originalCloudsImageMat.release();

				// increasing the image number
				imageNumber++;
				
				// if twelve images already processed, break the loop
				if (imageNumber > 12) {
					break;
				}
			}
		}

		// taking the time needed for algorithm to finish
		long end = System.currentTimeMillis();
		
		// printing the information to the console
		System.out.println("Algorithm lasted = " + (end - start));

		// preparing to store the collected data to the CSV file
		StringBuilder csvString = new StringBuilder("Time,");
		for (int i = 1; i <= data.size(); i++) {
			csvString.append(i);
			if (i <= data.size()) {
				csvString.append(",");
			}
		}

		// adding column data for the CSV file
		csvString.append("Sum");
		csvString.append("\n");

		// cloud counter initialization
		int maxNumberOfCloudMovement = 0;

		// getting the information about the cloud with the maximum number of movements
		// to adjust the number of clouds for the CSV file
		for (List<CoverageRecord> coverageRecordList : data) {
			if (coverageRecordList.size() > maxNumberOfCloudMovement) {
				maxNumberOfCloudMovement = coverageRecordList.size();
			}
		}

		// creating 2D array for CSV file structure
		String[][] dataArray = new String[data.size() + 1][maxNumberOfCloudMovement];

		// setting the coverage data to the 2D array
		for (int i = 0; i < data.size(); i++) {
			for (int j = 0; j < data.get(i).size(); j++) {
				if (dataArray[0][j] == null) {
					dataArray[0][j] = data.get(i).get(j).getTime().format(DateTimeFormatter.ISO_TIME);
				}
				dataArray[i + 1][j] = String.valueOf(data.get(i).get(j).getCoveragePercentage());
			}
		}

		// storing coverage data to the String buffer that will be 
		// persisted to the CSV file
		for (int i = 0; i < maxNumberOfCloudMovement; i++) {
			double coverageSum = 0;
			for (int j = 0; j < data.size() + 1; j++) {
				csvString.append(dataArray[j][i]);
				if (j > 0 && dataArray[j][i] != null) {
					coverageSum += Double.valueOf(dataArray[j][i]);
				}
				if (j <= data.size()) {
					csvString.append(",");
				}
			}
			csvString.append(coverageSum);
			csvString.append("\n");
		}

		// storing data to the CSV file
		PrintWriter csvPw = new PrintWriter(new File("data.csv"));
		csvPw.print(csvString);
		csvPw.flush();
		csvPw.close();
	}

	/**
	 * Returns the distance between two points.
	 * 
	 * @param firstPoint  first point coordinates
	 * @param secondPoint second point coordinates
	 * @return distance between the first and the second point
	 */
	private static double calculateDistance(Point firstPoint, Point secondPoint) {

		double differenceX = firstPoint.x - secondPoint.x;
		double differenceY = firstPoint.y - secondPoint.y;

		double distance = Math.sqrt(Math.pow(differenceX, 2) + Math.pow(differenceY, 2));

		return distance;
	}

	/**
	 * It contains data related to sun coverage record: ordinal number of the cloud
	 * contour on the image, time and coverage percentage.
	 * 
	 * @author Aleksander
	 *
	 */
	static class CoverageRecord {
		private int cloudOrdinal;
		private LocalTime time;
		private double coveragePercentage;

		/**
		 * Creates the coverage record object.
		 * 
		 * @param cloudOrdinal       - ordinal number of the cloud contour on the sky
		 *                           image
		 * @param time               - time of the day when the sun coverage is recorded
		 * @param coveragePercentage - percentage of the sun area covered by clouds
		 */
		public CoverageRecord(int cloudOrdinal, LocalTime time, double coveragePercentage) {
			super();
			this.cloudOrdinal = cloudOrdinal;
			this.time = time;
			this.coveragePercentage = coveragePercentage;
		}

		public double getCoveragePercentage() {
			return coveragePercentage;
		}

		public void setCoveragePercentage(double coveragePercentage) {
			this.coveragePercentage = coveragePercentage;
		}

		public int getCloudOrdinal() {
			return cloudOrdinal;
		}

		public void setCloudOrdinal(int cloudOrdinal) {
			this.cloudOrdinal = cloudOrdinal;
		}

		public LocalTime getTime() {
			return time;
		}

		public void setTime(LocalTime time) {
			this.time = time;
		}
	}
}
